/*
 * Copyright © 2022 Soren Stoutner <soren@stoutner.com>.
 *
 * This file is part of Privacy Browser PC <https://www.stoutner.com/privacy-browser-pc>.
 *
 * Privacy Browser PC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Privacy Browser PC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Privacy Browser PC.  If not, see <http://www.gnu.org/licenses/>.
 */

// Application headers.
#include "BrowserView.h"
#include "Settings.h"
#include "ui_BrowserView.h"
#include "filters/MouseEventFilter.h"
#include "helpers/DomainsDatabaseHelper.h"
#include "helpers/SearchEngineHelper.h"
#include "helpers/UserAgentHelper.h"
#include "interceptors/UrlRequestInterceptor.h"
#include "windows/BrowserWindow.h"

// Qt framework headers.
#include <QAction>

// Initialize the public static variables.
QString BrowserView::webEngineDefaultUserAgent = QStringLiteral("");

// Construct the class.
BrowserView::BrowserView(QWidget *parent) : QWidget(parent)
{
    // Initialize the variables.
    privacyWebEngineListPointer = new QList<PrivacyWebEngine*>;

    // Instantiate the browser view UI.
    Ui::BrowserView browserViewUi;

    // Setup the UI.
    browserViewUi.setupUi(this);

    // Get handles for the views.
    webEngineViewPointer = browserViewUi.webEngineView;

    // Create an off-the-record profile (the default when no profile name is specified).
    webEngineProfilePointer = new QWebEngineProfile(QStringLiteral(""));

    // Create a WebEngine page.
    webEnginePagePointer = new QWebEnginePage(webEngineProfilePointer);

    // Set the WebEngine page.
    webEngineViewPointer->setPage(webEnginePagePointer);

    // Get handles for the aspects of the WebEngine.
    webEngineHistoryPointer = webEnginePagePointer->history();
    webEngineSettingsPointer = webEngineViewPointer->settings();
    webEngineCookieStorePointer = webEngineProfilePointer->cookieStore();

    // Initialize the current privacy web engine pointer.
    currentPrivacyWebEnginePointer = new PrivacyWebEngine(webEngineViewPointer);

    // Populate the privacy web engine list.
    privacyWebEngineListPointer->append(currentPrivacyWebEnginePointer);

    // Set the cookie filter.
    webEngineCookieStorePointer->setCookieFilter([this](const QWebEngineCookieStore::FilterRequest &filterRequest)
    {
        // qDebug() << "Cookie page URL:  " << filterRequest.firstPartyUrl << ", Cookie URL:  " << filterRequest.origin << ",  Is third-party:  " << filterRequest.thirdParty;

        // Block all third party local storage requests, including the sneaky ones that don't register a first party URL.
        if (filterRequest.thirdParty || (filterRequest.firstPartyUrl == QStringLiteral("")))
            return false;

        // Check each tab to see if this local storage request should be allowed.
        for (PrivacyWebEngine *privacyWebEnginePointer : *privacyWebEngineListPointer)
        {
            // Allow this local storage request if it comes from a tab with local storage enabled.
            if (privacyWebEnginePointer->cookiesEnabled && (webEngineViewPointer->url().host() == filterRequest.firstPartyUrl.host()))
                return true;
        }

        // Block any remaining local storage requests.
        return false;
    });

    // Process cookie changes.
    connect(webEngineCookieStorePointer, SIGNAL(cookieAdded(QNetworkCookie)), this, SLOT(cookieAdded(QNetworkCookie)));
    connect(webEngineCookieStorePointer, SIGNAL(cookieRemoved(QNetworkCookie)), this, SLOT(cookieRemoved(QNetworkCookie)));

    // Store a copy of the WebEngine default user agent.
    webEngineDefaultUserAgent = webEngineProfilePointer->httpUserAgent();

    // Update the URL line edit when the URL changes.
    connect(webEngineViewPointer, SIGNAL(urlChanged(const QUrl)), this, SLOT(updateUrl(const QUrl)));

    // Update the progress bar.
    connect(webEngineViewPointer, SIGNAL(loadStarted()), this, SLOT(loadStarted()));
    connect(webEngineViewPointer, SIGNAL(loadProgress(const int)), this, SLOT(loadProgress(const int)));
    connect(webEngineViewPointer, SIGNAL(loadFinished(const bool)), this, SLOT(loadFinished()));

    // Instantiate the mouse event filter pointer.
    MouseEventFilter *mouseEventFilterPointer = new MouseEventFilter();

    // Install the mouse event filter.
    qApp->installEventFilter(mouseEventFilterPointer);

    // Process mouse forward and back commands.
    connect(mouseEventFilterPointer, SIGNAL(mouseBack()), this, SLOT(mouseBack()));
    connect(mouseEventFilterPointer, SIGNAL(mouseForward()), this, SLOT(mouseForward()));

    // Listen for hovered link URLs.
    connect(webEnginePagePointer, SIGNAL(linkHovered(const QString)), this, SLOT(pageLinkHovered(const QString)));

    // Instantiate the URL request interceptor.
    UrlRequestInterceptor *urlRequestInterceptorPointer = new UrlRequestInterceptor();

    // Set the URL request interceptor.
    webEngineProfilePointer->setUrlRequestInterceptor(urlRequestInterceptorPointer);

    // Reapply the domain settings when the host changes.
    connect(urlRequestInterceptorPointer, SIGNAL(applyDomainSettings(QString)), this, SLOT(applyDomainSettingsWithoutReloading(QString)));

    // Don't allow JavaScript to open windows.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptCanOpenWindows, false);

    // Allow keyboard navigation.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::SpatialNavigationEnabled, true);

    // Enable full screen support.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::FullScreenSupportEnabled, true);

    // Require user interaction to play media.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::PlaybackRequiresUserGesture, true);

    // Limit WebRTC to public IP addresses.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::WebRTCPublicInterfacesOnly, true);

    // Set the focus on the WebEngine view.
    webEngineViewPointer->setFocus();
}

BrowserView::~BrowserView()
{
    // Delay the deletion of the WebEngine page to prevent the following error:  `Release of profile requested but WebEnginePage still not deleted. Expect troubles !`
    webEnginePagePointer->deleteLater();
}

// The cookie is copied instead of referenced so that changes made to the cookie do not create a race condition with the display of the cookie in the dialog.
void BrowserView::addCookieToStore(QNetworkCookie cookie) const
{
    // Create a url.
    QUrl url;

    // Check to see if the domain does not start with a `.` because Qt makes this harder than it should be.  <https://doc.qt.io/qt-5/qwebenginecookiestore.html#setCookie>
    if (!cookie.domain().startsWith(QStringLiteral(".")))
    {
        // Populate the URL.
        url.setHost(cookie.domain());
        url.setScheme(QStringLiteral("https"));

        // Clear the domain from the cookie.
        cookie.setDomain(QStringLiteral(""));
    }

    // Add the cookie to the store.
    webEngineCookieStorePointer->setCookie(cookie, url);
}

void BrowserView::applyApplicationSettings()
{
    // Set the search engine URL.
    searchEngineUrl = SearchEngineHelper::getSearchUrl(Settings::searchEngine());

    // Emit the update search engine actions signal.
    emit updateSearchEngineActions(Settings::searchEngine());
}

// This exists as a separate function from `applyDomainSettings()` so it can be listed as a slot and function without the need for a boolean argument.
// Once <https://redmine.stoutner.com/issues/799> has been resolved this can be `const`.
void BrowserView::applyDomainSettingsAndReload()
{
    // Apply the domain settings.  `true` reloads the website.
    applyDomainSettings(webEngineViewPointer->url().host(), true);
}

// This exists as a separate function from `applyDomainSettings()` so it can be listed as a slot and function without the need for a boolean argument.
// Once <https://redmine.stoutner.com/issues/799> has been resolved this can be `const`.
void BrowserView::applyDomainSettingsWithoutReloading(const QString &hostname)
{
    // Apply the domain settings  `false` does not reload the website.
    applyDomainSettings(hostname, false);
}

// Once <https://redmine.stoutner.com/issues/799> has been resolved this can be `const`.
void BrowserView::applyDomainSettings(const QString &hostname, const bool reloadWebsite)
{
    // Get the record for the hostname.
    QSqlQuery domainQuery = DomainsDatabaseHelper::getDomainQuery(hostname);

    // Check if the hostname has domain settings.
    if (domainQuery.isValid())  // The hostname has domain settings.
    {
        // Get the domain record.
        QSqlRecord domainRecord = domainQuery.record();

        // Set the JavaScript status.
        switch (domainRecord.field(DomainsDatabaseHelper::JAVASCRIPT).value().toInt())
        {
            case (DomainsDatabaseHelper::SYSTEM_DEFAULT):
            {
                // Set the default JavaScript status.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, Settings::javaScriptEnabled());

                break;
            }

            case (DomainsDatabaseHelper::DISABLED):
            {
                // Disable JavaScript.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, false);

                break;
            }

            case (DomainsDatabaseHelper::ENABLED):
            {
                // Enable JavaScript.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, true);

                break;
            }
        }

        // Set the cookie status.  TODO.
        currentPrivacyWebEnginePointer->cookiesEnabled = Settings::cookiesEnabled();

        // Set DOM storage.
        switch (domainRecord.field(DomainsDatabaseHelper::DOM_STORAGE).value().toInt())
        {
            case (DomainsDatabaseHelper::SYSTEM_DEFAULT):
            {
                // Set the default DOM storage status.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::LocalStorageEnabled, Settings::domStorageEnabled());

                break;
            }

            case (DomainsDatabaseHelper::DISABLED):
            {
                // Disable DOM storage.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::LocalStorageEnabled, false);

                break;
            }

            case (DomainsDatabaseHelper::ENABLED):
            {
                // Enable DOM storage.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::LocalStorageEnabled, true);

                break;
            }
        }

        // Set the user agent.
        webEngineProfilePointer->setHttpUserAgent(UserAgentHelper::getResultingDomainSettingsUserAgent(domainRecord.field(DomainsDatabaseHelper::USER_AGENT).value().toString()));

        // Check if a custom zoom factor is set.
        if (domainRecord.field(DomainsDatabaseHelper::ZOOM_FACTOR).value().toInt())
        {
            // Store the current zoom factor.
            currentZoomFactor = domainRecord.field(DomainsDatabaseHelper::CUSTOM_ZOOM_FACTOR).value().toDouble();
        }
        else
        {
            // Reset the current zoom factor.
            currentZoomFactor = Settings::zoomFactor();
        }

        // Set the zoom factor.    The use of `currentZoomFactor` can be removed once <https://redmine.stoutner.com/issues/799> has been resolved.
        webEngineViewPointer->setZoomFactor(currentZoomFactor);

        // Apply the domain settings palette to the URL line edit.
        emit updateDomainSettingsIndicator(true, domainRecord.field(DomainsDatabaseHelper::DOMAIN_NAME).value().toString());
    }
    else  // The hostname does not have domain settings.
    {
        // Set the JavaScript status.
        webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, Settings::javaScriptEnabled());

        // Set the cookie status.
        currentPrivacyWebEnginePointer->cookiesEnabled = Settings::cookiesEnabled();

        // Set DOM storage.
        webEngineSettingsPointer->setAttribute(QWebEngineSettings::LocalStorageEnabled, Settings::domStorageEnabled());

        // Set the user agent.
        webEngineProfilePointer->setHttpUserAgent(UserAgentHelper::getUserAgentFromDatabaseName(Settings::userAgent()));

        // Store the current zoom factor.  This can be removed once <https://redmine.stoutner.com/issues/799> has been resolved.
        currentZoomFactor = Settings::zoomFactor();

        // Set the zoom factor.
        webEngineViewPointer->setZoomFactor(Settings::zoomFactor());

        // Apply the no domain settings palette to the URL line edit.
        emit updateDomainSettingsIndicator(false, QStringLiteral(""));
    }

    // Emit the update actions signals.
    emit updateJavaScriptAction(webEngineSettingsPointer->testAttribute(QWebEngineSettings::JavascriptEnabled));
    emit updateCookiesAction(currentPrivacyWebEnginePointer->cookiesEnabled);
    emit updateDomStorageAction(webEngineSettingsPointer->testAttribute(QWebEngineSettings::LocalStorageEnabled));
    emit updateUserAgentActions(webEngineProfilePointer->httpUserAgent());
    emit updateZoomFactorAction(webEngineViewPointer->zoomFactor());

    // Reload the website if requested.
    if (reloadWebsite)
    {
        webEngineViewPointer->reload();
    }
}

void BrowserView::applyOnTheFlySearchEngine(QAction *searchEngineActionPointer)
{
    // Store the search engine name.
    QString searchEngineName = searchEngineActionPointer->text();

    // Strip out any `&` characters.
    searchEngineName.remove('&');

    // Store the search engine string.
    searchEngineUrl = SearchEngineHelper::getSearchUrl(searchEngineName);
}

void BrowserView::applyOnTheFlyUserAgent(QAction *userAgentActionPointer) const
{
    // Get the user agent name.
    QString userAgentName = userAgentActionPointer->text();

    // Strip out any `&` characters.
    userAgentName.remove('&');

    // Apply the user agent.
    webEngineProfilePointer->setHttpUserAgent(UserAgentHelper::getUserAgentFromTranslatedName(userAgentName));

    // Reload the website.
    webEngineViewPointer->reload();
}

// This can be const once <https://redmine.stoutner.com/issues/799> has been resolved.
void BrowserView::applyOnTheFlyZoomFactor(const double &zoomFactor)
{
    // Update the current zoom factor.  This can be removed once <https://redmine.stoutner.com/issues/799> has been resolved.
    currentZoomFactor = zoomFactor;

    // Set the zoom factor.
    webEngineViewPointer->setZoomFactor(zoomFactor);
}

void BrowserView::back() const
{
    // Go back.
    webEngineViewPointer->back();
}

void BrowserView::cookieAdded(const QNetworkCookie &cookie) const
{
    // Add the cookie to the cookie list.
    emit addCookie(cookie);
}

void BrowserView::cookieRemoved(const QNetworkCookie &cookie) const
{
    // Remove the cookie from the cookie list.
    emit removeCookie(cookie);
}

void BrowserView::deleteAllCookies() const
{
    // Delete all the cookies.
    webEngineCookieStorePointer->deleteAllCookies();
}

void BrowserView::deleteCookieFromStore(const QNetworkCookie &cookie) const
{
    // Delete the cookie.
    webEngineCookieStorePointer->deleteCookie(cookie);
}

void BrowserView::forward() const
{
    // Go forward.
    webEngineViewPointer->forward();
}

void BrowserView::home() const
{
    // Load the homepage.
    webEngineViewPointer->load(QUrl::fromUserInput(Settings::homepage()));
}

void BrowserView::loadFinished() const
{
    // Hide the progress bar.
    emit hideProgressBar();
}

void BrowserView::loadInitialWebsite()
{
    // Apply the application settings.
    applyApplicationSettings();

    // Get the arguments.
    QStringList argumentsStringList = qApp->arguments();

    // Check to see if the arguments lists contains a URL.
    if (argumentsStringList.size() > 1)
    {
        // Load the URL from the arguments list.
        webEngineViewPointer->load(QUrl::fromUserInput(argumentsStringList.at(1)));
    }
    else
    {
        // Load the homepage.
        home();
    }
}

void BrowserView::loadProgress(const int &progress) const
{
    // Show the progress bar.
    emit showProgressBar(progress);
}

void BrowserView::loadStarted() const
{
    // Show the progress bar.
    emit showProgressBar(0);
}

void BrowserView::loadUrlFromLineEdit(QString url) const
{
    // Decide if the text is more likely to be a URL or a search.
    if (url.startsWith("file://"))  // The text is likely a file URL.
    {
        // Load the URL.
        webEngineViewPointer->load(QUrl::fromUserInput(url));
    }
    else if (url.contains("."))  // The text is likely a URL.
    {
        // Check if the URL does not start with a valid protocol.
        if (!url.startsWith("http"))
        {
            // Add `https://` to the beginning of the URL.
            url = "https://" + url;
        }

        // Load the URL.
        webEngineViewPointer->load(QUrl::fromUserInput(url));
    }
    else  // The text is likely a search.
    {
        // Load the search.
        webEngineViewPointer->load(QUrl::fromUserInput(searchEngineUrl + url));
    }
}

void BrowserView::mouseBack() const
{
    // Go back if possible.
    if (webEngineHistoryPointer->canGoBack())
    {
        // Clear the URL line edit focus.
        emit clearUrlLineEditFocus();

        // Go back.
        webEngineViewPointer->back();
    }
}

void BrowserView::mouseForward() const
{
    // Go forward if possible.
    if (webEngineHistoryPointer->canGoForward())
    {
        // Clear the URL line edit focus.
        emit clearUrlLineEditFocus();

        // Go forward.
        webEngineViewPointer->forward();
    }
}

void BrowserView::pageLinkHovered(const QString &linkUrl) const
{
    // Emit a signal so that the browser window can update the status bar.
    emit linkHovered(linkUrl);
}

void BrowserView::refresh() const
{
    // Reload the website.
    webEngineViewPointer->reload();
}

void BrowserView::toggleCookies()
{
    // Toggle cookies.
    currentPrivacyWebEnginePointer->cookiesEnabled = !currentPrivacyWebEnginePointer->cookiesEnabled;

    // Update the cookies icon.
    emit updateCookiesAction(currentPrivacyWebEnginePointer->cookiesEnabled);

    // Reload the website.
    webEngineViewPointer->reload();
}

void BrowserView::toggleJavaScript() const
{
    // Toggle JavaScript.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, !webEngineSettingsPointer->testAttribute(QWebEngineSettings::JavascriptEnabled));

    // Update the JavaScript icon.
    emit updateJavaScriptAction(webEngineSettingsPointer->testAttribute(QWebEngineSettings::JavascriptEnabled));

    // Reload the website.
    webEngineViewPointer->reload();
}

void BrowserView::toggleDomStorage() const
{
    // Toggle DOM storage.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::LocalStorageEnabled, !webEngineSettingsPointer->testAttribute(QWebEngineSettings::LocalStorageEnabled));

    // Update the DOM storage action icon.
    emit updateDomStorageAction(webEngineSettingsPointer->testAttribute(QWebEngineSettings::LocalStorageEnabled));

    // Reload the website.
    webEngineViewPointer->reload();
}

void BrowserView::updateUrl(const QUrl &url) const
{
    // Update the URL line edit.
    emit updateUrlLineEdit(url);

    // Update the status of the forward and back buttons.
    emit updateBackAction(webEngineHistoryPointer->canGoBack());
    emit updateForwardAction(webEngineHistoryPointer->canGoForward());

    // Reapply the zoom factor.  This is a bug in QWebEngineView that resets the zoom with every load.  <https://redmine.stoutner.com/issues/799>
    webEngineViewPointer->setZoomFactor(currentZoomFactor);
}
